#! /usr/bin/perl
# Working Time Follow up
# (c) jan@jan-trippler.de 2016
# Tool to interactively create a work time usage protocol
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

use strict;
use warnings;
use Tk;
use utf8;
use Cwd qw(abs_path);
use File::Basename;
use Date::Calc qw(Delta_DHMS Day_of_Week_to_Text);

##### environment settings #####
my $bin_path  = abs_path(dirname($0));
my $data_path = abs_path($bin_path . "/../data");

# show licence information on STDERR
my @show = <DATA>;
print STDERR @show, "\n";
close DATA;

# if requested show gpl and exit
if (defined $ARGV[0] && $ARGV[0] eq '-gpl') {
  open GPL, "$bin_path/gpl-2.0.txt" or die "can't open gpl-2.0.txt\n";
  my @gpl = <GPL>;
  print @gpl;
  close GPL;
  exit 0;
}

##### constants #####
my @weekdays = qw(Sonntag Montag Dienstag Mittwoch Donnerstag Freitag Samstag);

##### configuration settings #####
my %cfg = ('display' => {},
           'gui'     => {}
);
# GUI settings
$cfg{'gui'}->{'col_frames'}   = '#9090e0';
$cfg{'gui'}->{'col_edit'}     = '#b0b0ff';
$cfg{'gui'}->{'col_head'}     = '#7070e0';
$cfg{'gui'}->{'col_msg'}      = '#20d080';
$cfg{'gui'}->{'col_warn'}     = '#a04040';
$cfg{'gui'}->{'col_disabled'} = '#6060d0';
$cfg{'gui'}->{'col_font'}     = '#6060d0';
$cfg{'gui'}->{'lang'}         = 'de';
# jake
$cfg{'gui'}->{'screen_pos'}   = '+1660+10';
$cfg{'gui'}->{'padx'}         = 2;
$cfg{'gui'}->{'pady'}         = 0;
$cfg{'gui'}->{'grid_padx'}    = 2;
$cfg{'gui'}->{'grid_pady'}    = 2;
$cfg{'gui'}->{'wide_padx'}    = 5;
$cfg{'gui'}->{'wide_pady'}    = 3;
$cfg{'gui'}->{'tree_height'}  = 36;
$cfg{'gui'}->{'sel_height'}   = 10;
$cfg{'gui'}->{'stxt_height'}  = 5;
$cfg{'gui'}->{'ltxt_height'}  = 8;

##### named ui elements
my $restButton   = undef;
my $startButton  = undef;
my $finishButton = undef;
my $worktime_lbl = undef;
my $frm_text     = undef;

##### global variables bound to a ui element
my $curr_date     = '';  # current date
my $curr_time     = '';  # current time
my $day_diff      = '';  # time since program start
my $work_diff     = '';  # worked time since program start
my $job_diff      = '';  # worked in current job
my $curr_job      = '';  # current job
my $curr_activity = '';  # current activity (comment)
my $new_job       = '';  # entry for a new job

##### global parameters
my $activity_file = '';  # where to save job times to
my $prot_date     = '';  # date of program start is used as job date
my $is_rest       = 0;   # flag if I'm resting
my $is_finished   = 0;   # flag if I'm working
my @jobs          = ();  # list of currently known jobs
# calculating times
my %times = (
  'all' => {},
  'job' => {}
);
# daily time since start
$times{'all'}->{'started'}    = time;
# time of current working period start
$times{'all'}->{'work_start'} = $times{'all'}->{'started'};
# daily work time
$times{'all'}->{'worked'}     = 0;
# start time in current job (from start or since last pause)
$times{'job'}->{'started'}    = $times{'all'}->{'started'};
# work time in current job
$times{'job'}->{'worked'}     = 0;

##### helper functions #####

# format a difference between two times
# param - difference between 2 timestamps in seconds
# returns string in the form hh:mm:ss
sub format_time_diff {
  my $diff = shift;
  my $hrs  = int($diff / 3600);
  my $mins = int(($diff - $hrs * 3600) / 60);
  my $secs = $diff - $hrs * 3600 - $mins * 60;
  return sprintf "%02d:%02d:%02d", $hrs, $mins, $secs;
}

# format a timestamp into different output formats
# param - timestamp
# returns an array of strings:
#   wdayname, dd.MM.yyyy
#   yyyy-MM-dd
#   yyyy-MM
#   hh:mm:ss
sub format_date_time {
  my $ts = shift;
  my ($sec, $min, $hrs, $mday, $mon, $year, $wday, @dummy) = localtime($ts);
  $mon  += 1;
  $year += 1900;
  my @rc = (
    (sprintf "%s, %02d.%02d.%04d", $weekdays[$wday], $mday, $mon, $year),
    (sprintf "%04d-%02d-%02d", $year, $mon, $mday),
    (sprintf "%04d-%02d", $year, $mon),
    (sprintf "%02d:%02d:%02d", $hrs, $min, $sec)
  );
  return @rc;
}

# compute current times, is called every second
sub new_time {
  my $this_time  = time;
  my @dt_strings = format_date_time($this_time);
  $curr_date     = $dt_strings[0];
  $curr_time     = $dt_strings[3];
  $day_diff      = format_time_diff($this_time - $times{'all'}->{'started'});
  if ($is_rest == 0) {
    $work_diff = format_time_diff($this_time - $times{'all'}->{'work_start'} + $times{'all'}->{'worked'});
    $job_diff  = format_time_diff($this_time - $times{'all'}->{'work_start'} + $times{'job'}->{'worked'});
  }
  if ($times{'all'}->{'worked'} >= 28800) {
    $worktime_lbl->configure(
      -foreground => $cfg{'gui'}->{'col_warn'}
    );
  }
}

# load jobs from file
sub load_jobs {
  if (-f "$data_path/jobs.txt") {
    open JOB, "$data_path/jobs.txt" or die "can't open <$data_path/jobs.txt> ($!)";
    @jobs = <JOB>;
    close JOB;
    for (my $i = 0; $i <= $#jobs; $i++) {
      $jobs[$i] =~ s/\s*\r?\n//;
    }
  }
}

# save jobs to file
sub store_jobs {
  if (-f "$data_path/jobs.txt") {
    open JOB, ">$data_path/jobs.txt" or die "can't open <$data_path/jobs.txt> ($!)";
    $| = 1; # Before writing!
    print JOB join("\n", @jobs);
    close JOB;
    JOB->autoflush(1);
  }
}

# change status from work to pause
sub work2pause {
  # add time since last start to working times
  if ($is_rest == 0) {
    $times{'job'}->{'worked'} += time - $times{'all'}->{'work_start'};
    $times{'all'}->{'worked'} += time - $times{'all'}->{'work_start'};
  }
  $is_rest = 1;
}

# change status from pause to work
sub pause2work {
  # set new work start time
  if ($is_rest == 1) {
    $times{'all'}->{'work_start'} = time;
  }
  $is_rest = 0;
}

# save last activity
sub store_activity {
  unless (-f $activity_file) {
    open ACT, ">$activity_file" or die "can't open <$activity_file> ($!)";
    print ACT "Tag|Start|Ende|Dauer|Job|Aktivität\n";
    close ACT;
  }
  work2pause();
  my $prot_job_diff = format_time_diff(time - $times{'all'}->{'work_start'} + $times{'job'}->{'worked'});
  my (@fmt_times) = format_date_time($times{'job'}->{'started'});
  my $prot_start_job = $fmt_times[3];
  (@fmt_times) = format_date_time(time);
  my $prot_end_job = $fmt_times[3];
  open ACT, ">>$activity_file" or die "can't open <$activity_file> ($!)";
  print ACT "$prot_date|$prot_start_job|$prot_end_job|$job_diff|$curr_job|$curr_activity\n";
  close ACT;
}

# change activity
sub change_activity {
  my $job = shift;
  store_activity();
  $curr_activity = '' unless $job eq $curr_job;
  $curr_job      = $job;
  pause2work();
  $times{'job'}->{'started'} = time;
  $times{'job'}->{'worked'}  = 0;
}

# load jobs into GUI
sub jobs_to_gui {
  # first job == default if no current job exists
  if ($curr_job eq '') {
    my $fst_job = '';
    $fst_job = $jobs[0] if $#jobs >= 0;
    $curr_job = $fst_job;
  }
  # existing jobs
  for(my $job_cnt = 0; $job_cnt <= $#jobs; $job_cnt++) {
    my $job = $jobs[$job_cnt];
    $frm_text->Label(
      -text => $job,
      -padx => $cfg{'gui'}->{'padx'},
      -background => $cfg{'gui'}->{'col_frames'}
    )->grid(
      -column => 1,
      -row => $job_cnt,
      -padx => $cfg{'gui'}->{'grid_padx'},
      -pady => $cfg{'gui'}->{'grid_pady'},
      -sticky  => 'nsw'
    );
    $frm_text->Button(
      -relief => 'flat',
      -background => $cfg{'gui'}->{'col_edit'},
      -borderwidth => 1,
      -disabledforeground => $cfg{'gui'}->{'col_enabled'},
      -foreground => $cfg{'gui'}->{'col_msg'},
      -padx => $cfg{'gui'}->{'padx'},
      -pady => $cfg{'gui'}->{'pady'},
      -text => "   >>   ",
      -command => sub {
          change_activity($job);
          $restButton->configure(
            -state => 'normal',
            -background => $cfg{'gui'}->{'col_frames'}
          );
          $startButton->configure(
            -state => 'disabled',
            -background => $cfg{'gui'}->{'col_frames'},
            -disabledforeground => $cfg{'gui'}->{'col_disabled'}
          );
        }
    )->grid(
      -column => 2,
      -row => $job_cnt,
      -padx => $cfg{'gui'}->{'grid_padx'},
      -pady => $cfg{'gui'}->{'grid_pady'},
      -sticky  => 'nse'
    )->configure(
      -state => 'normal',
      -background => $cfg{'gui'}->{'col_frames'}
    );
  }
}


##### MAIN: GET DATA AND CREATE GUI #####

my @start_data = format_date_time(time);
my $start_time = $start_data[3];
$curr_date     = $start_data[0];
$curr_time     = $start_data[3];
$activity_file = sprintf("%s/%s.csv", $data_path, $start_data[2]);
$prot_date     = $start_data[1];
load_jobs();

# the main window
my $win_main = MainWindow->new(-background => $cfg{'gui'}->{'col_frames'});
$win_main->title('WTF! Working Time Follow up by software@jan-trippler.de');
$win_main->geometry($cfg{'gui'}->{'screen_pos'});
# refresh every second
$win_main->repeat(1000 => \&new_time);

# toplevel frames:
# status (times left, date)
my $frm_status =
  $win_main->Frame(
    -background => $cfg{'gui'}->{'col_frames'}
  );
# start/pause/finish buttons
my $frm_buttons =
  $win_main->Frame(
    -background => $cfg{'gui'}->{'col_frames'}
  );
# current activity
my $frm_activity =
  $win_main->Frame(
    -background => $cfg{'gui'}->{'col_frames'}
  );
# jobs
$frm_text =
  $win_main->Frame(
    -background => $cfg{'gui'}->{'col_frames'}
  );
# new job
my $frm_insert =
  $win_main->Frame(
    -background => $cfg{'gui'}->{'col_frames'}
  );
# pack the frames which are visible at startup
$frm_status->pack(
  -side => 'top',
  -fill => 'both'
);
$frm_buttons->pack(
  -side => 'top',
  -fill => 'both'
);
$frm_activity->pack(
  -side => 'top',
  -fill => 'both'
);
$frm_text->pack(
  -side => 'top',
  -fill => 'both'
);
$frm_insert->pack(
  -side => 'bottom',
  -fill => 'both'
);
# the elements of the status frame: date, current time, 
# day's work time
$frm_status->Label(
  -textvariable => \$curr_date,
  -padx => $cfg{'gui'}->{'padx'},
  -background => $cfg{'gui'}->{'col_frames'}
)->pack(
  -side => 'left',
  -anchor => 'w'
);
$frm_status->Label(
  -padx => $cfg{'gui'}->{'padx'},
  -background => $cfg{'gui'}->{'col_frames'},
  -textvariable => \$curr_time,
  -foreground => $cfg{'gui'}->{'col_msg'}
)->pack(
  -side => 'left',
  -anchor => 'w',
  -fill => 'x',
  -expand => 1
);
$frm_status->Label(
  -text => "Gestartet: ",
  -padx => $cfg{'gui'}->{'padx'},
  -background => $cfg{'gui'}->{'col_frames'}
)->pack(
  -side => 'left',
  -anchor => 'w'
);
$frm_status->Label(
  -padx => $cfg{'gui'}->{'padx'},
  -background => $cfg{'gui'}->{'col_frames'},
  -text => $start_time
)->pack(
  -side => 'left',
  -anchor => 'w',
  -fill => 'x',
  -expand => 1
);
$frm_status->Label(
  -text => "Verronnen: ",
  -padx => $cfg{'gui'}->{'padx'},
  -background => $cfg{'gui'}->{'col_frames'}
)->pack(
  -side => 'left',
  -anchor => 'w'
);
$frm_status->Label(
  -padx => $cfg{'gui'}->{'padx'},
  -background => $cfg{'gui'}->{'col_frames'},
  -textvariable => \$day_diff,
  -foreground => $cfg{'gui'}->{'col_msg'}
)->pack(
  -side => 'left',
  -anchor => 'w',
  -fill => 'x',
  -expand => 1
);
$frm_status->Label(
  -text => "Produktiv: ",
  -padx => $cfg{'gui'}->{'padx'},
  -background => $cfg{'gui'}->{'col_frames'}
)->pack(
  -side => 'left',
  -anchor => 'w'
);
$worktime_lbl = $frm_status->Label(
  -padx => $cfg{'gui'}->{'padx'},
  -background => $cfg{'gui'}->{'col_frames'},
  -textvariable => \$work_diff,
  -foreground => $cfg{'gui'}->{'col_msg'}
)->pack(
  -side => 'left',
  -anchor => 'w',
  -fill => 'x',
  -expand => 1
);
# the action buttons
$restButton = $frm_buttons->Button(
  -relief => 'flat',
  -background => $cfg{'gui'}->{'col_edit'},
  -borderwidth => 1,
  -disabledforeground => $cfg{'gui'}->{'col_enabled'},
  -padx => $cfg{'gui'}->{'padx'},
  -pady => $cfg{'gui'}->{'pady'},
  -text => ' Pause ',
  -command => sub {
      work2pause();
      $startButton->configure(
        -state => 'normal',
        -background => $cfg{'gui'}->{'col_frames'}
      );
      $restButton->configure(
        -state => 'disabled',
        -background => $cfg{'gui'}->{'col_frames'},
        -disabledforeground => $cfg{'gui'}->{'col_disabled'}
      );
    }
)->grid(
  -column => 0,
  -row => 0,
  -padx => $cfg{'gui'}->{'grid_padx'},
  -pady => $cfg{'gui'}->{'grid_pady'},
  -sticky  => 'nsew'
);
$restButton->configure(
  -state => 'normal',
  -background => $cfg{'gui'}->{'col_frames'}
);
$startButton = $frm_buttons->Button(
  -relief => 'flat',
  -background => $cfg{'gui'}->{'col_edit'},
  -borderwidth => 1,
  -disabledforeground => $cfg{'gui'}->{'col_enabled'},
  -padx => $cfg{'gui'}->{'padx'},
  -pady => $cfg{'gui'}->{'pady'},
  -text => " Weiter geht's ",
  -command => sub {
      pause2work();
      $restButton->configure(
        -state => 'normal',
        -background => $cfg{'gui'}->{'col_frames'}
      );
      $startButton->configure(
        -state => 'disabled',
        -background => $cfg{'gui'}->{'col_frames'},
        -disabledforeground => $cfg{'gui'}->{'col_disabled'}
      );
    }
)->grid(
  -column => 1,
  -row => 0,
  -padx => $cfg{'gui'}->{'grid_padx'},
  -pady => $cfg{'gui'}->{'grid_pady'},
  -sticky  => 'nsew'
);
$startButton->configure(
  -state => 'disabled',
  -background => $cfg{'gui'}->{'col_frames'},
  -disabledforeground => $cfg{'gui'}->{'col_disabled'}
);
$finishButton = $frm_buttons->Button(
  -relief => 'flat',
  -background => $cfg{'gui'}->{'col_edit'},
  -borderwidth => 1,
  -disabledforeground => $cfg{'gui'}->{'col_enabled'},
  -padx => $cfg{'gui'}->{'padx'},
  -pady => $cfg{'gui'}->{'pady'},
  -text => " Feierahmd! ",
  -command => sub {
      store_activity();
      $is_finished = 1;
      exit 0;
    }
)->grid(
  -column => 2,
  -row => 0,
  -padx => $cfg{'gui'}->{'grid_padx'},
  -pady => $cfg{'gui'}->{'grid_pady'},
  -sticky  => 'nsew'
);
$finishButton->configure(
  -state => 'normal',
  -background => $cfg{'gui'}->{'col_frames'}
);
# the activity elements
$frm_activity->Label(
  -text => "Job: ",
  -padx => $cfg{'gui'}->{'padx'},
  -background => $cfg{'gui'}->{'col_frames'}
)->grid(
  -column => 0,
  -row => 0,
  -padx => $cfg{'gui'}->{'grid_padx'},
  -pady => $cfg{'gui'}->{'grid_pady'},
  -sticky  => 'nse'
);
$frm_activity->Label(
  -textvariable => \$curr_job,
  -padx => $cfg{'gui'}->{'padx'},
  -background => $cfg{'gui'}->{'col_frames'}
)->grid(
  -column => 1,
  -row => 0,
  -padx => $cfg{'gui'}->{'grid_padx'},
  -pady => $cfg{'gui'}->{'grid_pady'},
  -sticky  => 'nsw'
);
$frm_activity->Label(
  -text => 'im Job verstrichen',
  -padx => $cfg{'gui'}->{'padx'},
  -background => $cfg{'gui'}->{'col_frames'}
)->grid(
  -column => 2,
  -row => 0,
  -padx => $cfg{'gui'}->{'grid_padx'},
  -pady => $cfg{'gui'}->{'grid_pady'},
  -sticky  => 'nsw'
);
$frm_activity->Label(
  -textvariable => \$job_diff,
  -padx => $cfg{'gui'}->{'padx'},
  -foreground => $cfg{'gui'}->{'col_msg'},
  -background => $cfg{'gui'}->{'col_frames'}
)->grid(
  -column => 3,
  -row => 0,
  -padx => $cfg{'gui'}->{'grid_padx'},
  -pady => $cfg{'gui'}->{'grid_pady'},
  -sticky  => 'nsw'
);
$frm_activity->Label(
  -text => 'Aktivität immer vor dem Wechsel in den nächsten Job eintragen!',
  -padx => $cfg{'gui'}->{'padx'},
  -background => $cfg{'gui'}->{'col_frames'}
)->grid(
  -column => 0,
  -row => 1,
  -columnspan => 4,
  -padx => $cfg{'gui'}->{'grid_padx'},
  -pady => $cfg{'gui'}->{'grid_pady'},
  -sticky  => 'ns'
);
$frm_activity->Label(
  -text => "Was mach ich denn da? ",
  -padx => $cfg{'gui'}->{'padx'},
  -background => $cfg{'gui'}->{'col_frames'}
)->grid(
  -column => 0,
  -row => 2,
  -padx => $cfg{'gui'}->{'grid_padx'},
  -pady => $cfg{'gui'}->{'grid_pady'},
  -sticky  => 'nse'
);
$frm_activity->Entry(
  -relief => 'flat',
  -width => 80,
  -background => $cfg{'gui'}->{'col_edit'},
  -foreground => $cfg{'gui'}->{'col_font'},
  -textvariable => \$curr_activity
)->grid(
  -column => 1,
  -row => 2,
  -columnspan => 3,
  -padx => $cfg{'gui'}->{'grid_padx'},
  -pady => $cfg{'gui'}->{'grid_pady'},
  -sticky  => 'nse'
);
# load jobs
jobs_to_gui();
# new job elements
$frm_insert->Label(
  -text => "Was Neues ...",
  -padx => $cfg{'gui'}->{'padx'},
  -background => $cfg{'gui'}->{'col_frames'}
)->grid(
  -column => 0,
  -row => 0,
  -padx => $cfg{'gui'}->{'grid_padx'},
  -pady => $cfg{'gui'}->{'grid_pady'},
  -sticky  => 'nsw'
);
$frm_insert->Entry(
  -relief => 'flat',
  -width => 80,
  -background => $cfg{'gui'}->{'col_edit'},
  -foreground => $cfg{'gui'}->{'col_font'},
  -textvariable => \$new_job
)->grid(
  -column => 1,
  -row => 0,
  -padx => $cfg{'gui'}->{'grid_padx'},
  -pady => $cfg{'gui'}->{'grid_pady'},
  -sticky  => 'nsw'
);
$frm_insert->Button(
  -relief => 'flat',
  -background => $cfg{'gui'}->{'col_edit'},
  -borderwidth => 1,
  -disabledforeground => $cfg{'gui'}->{'col_enabled'},
  -foreground => $cfg{'gui'}->{'col_msg'},
  -padx => $cfg{'gui'}->{'padx'},
  -pady => $cfg{'gui'}->{'pady'},
  -text => " der auch noch! ",
  -command => sub {
      push(@jobs, $new_job);
      store_jobs();
      jobs_to_gui();
      $new_job = '';
    }
)->grid(
  -column => 2,
  -row => 0,
  -padx => $cfg{'gui'}->{'grid_padx'},
  -pady => $cfg{'gui'}->{'grid_pady'},
  -sticky  => 'nse'
)->configure(
  -state => 'normal',
  -background => $cfg{'gui'}->{'col_frames'}
);

##### now execute the application #####

MainLoop;

# and now we're finished
exit 0;

__DATA__

wtf.pl version 0.1, Copyright (C) 2016 Jan Trippler
wtf.pl comes with ABSOLUTELY NO WARRANTY. This is free 
software, and you are welcome to redistribute it under certain
conditions; type `wtf.pl -gpl' or have a look at the file 
gpl-2.0.txt for details.
